/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package org.hakret.toschu.turingmaschine;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 *
 * @author toschu
 */
public class TuringCompiler {

    private String input;
    private Document document;

    public TuringCompiler(String input) {
        this.input = input;
    }

    public void buildXMLFile(String ouputFileName)
            throws ParserConfigurationException, TransformerException, TransformerConfigurationException, IOException {
        initDocument();
        input = input.replace(" ", "");
        input = input.replace("\t", "");
        input = input.replace("\n", "");
        input = input.replace("\r", "");
        String transition = input.split("00")[0];
        String finalStates = input.split("00")[1];
        List<String[]> transistions = listTransitions(transition);
        for (String[] currentTransistion : transistions) {
            addTransition(currentTransistion);
        }
        setFinalStates(finalStates);
        addEnd();
        printDoc(document, ouputFileName);

    }

    public void setFinalStates(String finalStates) {
        String[] finalStatesNumbers = finalStates.split("0");
        for (String currentState : finalStatesNumbers) {
            Element currentElement = getState(currentState);
            if (currentElement != null) {
                currentElement.setAttribute("finalstate", "true");
            }
        }
    }

    public Element getState(String name) {
        NodeList childNodes = document.getElementsByTagName("STATE");
        Element result = null;
        int number = countOnes(name) - 1;
        name = "q_" + number;
        for (int i = 0; i < childNodes.getLength(); i++) {
            Node currentNode = childNodes.item(i);
            if (currentNode.getNodeType() == Node.ELEMENT_NODE) {
                Element eElement = (Element) currentNode;
                if (eElement.getAttribute("name").equals(name)) {
                    result = eElement;
                    break;
                }
            }
        }
        return result;
    }

    public void addTransition(String[] content) {
        Element currentState = getState(content[0]);
        if (currentState == null) {
            currentState = addState(content[0], false);
        }
        Element transition = document.createElement("TRANSITION");
        String targetstate = "q_" + (countOnes(content[2]) - 1);
        Element targetState = getState(content[2]);
        if (targetState == null) {
            targetState = addState(content[2], false);
        }
        transition.setAttribute("target", targetstate);
        currentState.appendChild(transition);
        Element label = document.createElement("LABEL");
        String read = "" + (countOnes(content[1]) - 1);
        label.setAttribute("read", read);
        String write = "" + (countOnes(content[3]) - 1);
        label.setAttribute("write", write);
        int directionCount = countOnes(content[4]);
        String diretcion = "NONE";
        switch (directionCount) {
            case 1:
                diretcion = "LEFT";
                break;
            case 2:
                diretcion = "RIGHT";
                break;
            default:
                diretcion = "NONE";
                break;
        }
        label.setAttribute("direction", diretcion);
        transition.appendChild(label);
    }

    public Element addState(String name, boolean finalState) {
        Element state = document.createElement("STATE");
        state.setAttribute("name", "q_" + (countOnes(name) - 1));
        if (finalState) {
            state.setAttribute("finalstate", "true");
        } else {
            state.setAttribute("finalstate", "false");
        }
        document.getFirstChild().appendChild(state);
        return state;
    }

    public List<String[]> listTransitions(String content) {
        String[] tansitions = content.split("0");
        //System.out.println(Arrays.deepToString(tansitions));
        List<String[]> tansitionParts = new ArrayList<>();
        String[] oneTransition = new String[5];
        for (int i = 0; i < tansitions.length; i++) {
            oneTransition[0] = tansitions[i];
            oneTransition[1] = tansitions[i + 1];
            oneTransition[2] = tansitions[i + 2];
            oneTransition[3] = tansitions[i + 3];
            oneTransition[4] = tansitions[i + 4];
            tansitionParts.add(oneTransition);
            oneTransition = new String[5];
            i += 4;
        }
        return tansitionParts;
    }

    public int countOnes(String text) {
        int result = text.length() - text.replaceAll("1", "").length();
        return result;
    }

    public void addEnd() {
        Element initialState = document.createElement("INITIALSTATE");
        initialState.setAttribute("value", "q_0");
        document.getFirstChild().appendChild(initialState);
        Element tpInitialChar = document.createElement("TAPEINITIALCHAR");
        tpInitialChar.setAttribute("value", "$");
        document.getFirstChild().appendChild(tpInitialChar);
    }

    public void initDocument() throws ParserConfigurationException {
        DocumentBuilderFactory dbFactory
                = DocumentBuilderFactory.newInstance();
        DocumentBuilder dBuilder = dbFactory.newDocumentBuilder();
        document = dBuilder.newDocument();
        Element root = document.createElement("AUTOMATON");
        Element type = document.createElement("TYPE");
        type.setAttribute("value", "TM");
        root.appendChild(type);
        document.appendChild(root);
        Element alphabet = document.createElement("ALPHABET");
        root.appendChild(alphabet);
        Element zeroSymbol = document.createElement("ITEM");
        zeroSymbol.setAttribute("value", "0");
        Element oneSymbol = document.createElement("ITEM");
        oneSymbol.setAttribute("value", "1");
        alphabet.appendChild(zeroSymbol);
        alphabet.appendChild(oneSymbol);
        Element tapeAlphabet = document.createElement("TAPEALPHABET");
        Element tpdefaultSymbol = document.createElement("ITEM");
        tpdefaultSymbol.setAttribute("value", "$");
        Element tpzeroSymbol = document.createElement("ITEM");
        tpzeroSymbol.setAttribute("value", "0");
        Element tponeSymbol = document.createElement("ITEM");
        tponeSymbol.setAttribute("value", "1");
        tapeAlphabet.appendChild(tpdefaultSymbol);
        tapeAlphabet.appendChild(tpzeroSymbol);
        tapeAlphabet.appendChild(tponeSymbol);
        root.appendChild(tapeAlphabet);

    }

    public void printDoc(Document document, String outputFileName)
            throws TransformerConfigurationException, TransformerException, IOException {
        // write the content into xml file
        TransformerFactory transformerFactory = TransformerFactory.newInstance();
        Transformer transformer = transformerFactory.newTransformer();
        DOMSource source = new DOMSource(document);
        StreamResult result = new StreamResult(System.out);
        File outFile = new File(outputFileName);
        if (outFile.exists()) {
            outFile.delete();

        }
        outFile.createNewFile();
        StreamResult result2 = new StreamResult(outFile);
        transformer.transform(source, result);
        transformer.transform(source, result2);
    }

    public String getInput() {
        return input;
    }

    public void setInput(String input) {
        this.input = input;
    }

    public Document getDocument() {
        return document;
    }

    public void setDocument(Document document) {
        this.document = document;
    }

}
